# -*- coding: utf-8 -*-
# Copyright 2010-2013 Kolab Systems AG (http://www.kolabsys.com)
#
# Jeroen van Meeuwen (Kolab Systems) <vanmeeuwen a kolabsys.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

from pykolab.translate import _
from os.path import isfile, isdir, join
import os, sys

def import_object(module_name, attr):
    """
        Import from the module with the given name the object with the given
        attribute name.
    """
    module = __import__(module_name, level=0)
    obj = module
    for name in module_name.split(".")[1:]:
        obj = getattr(obj, name)
    try:
        return getattr(obj, attr)
    except AttributeError:
        raise ImportError, "cannot import name %s" % attr

def list_commands(commands, *args, **kw):
    """
        List commands or components.
    """
    show_commands(group_commands(commands))

def group_commands(commands):
    """
        Return the grouped commands.
    """
    grouped_commands = {}

    for command, definition in commands.items():
        if isinstance(command, tuple):
            command_group, command = command
            grouped_commands[command_group] = {command: definition}
        else:
            grouped_commands[command] = definition

    return grouped_commands

def get_sorted_commands(commands):
    """
        Return the sorted keys of the commands dictionary.
    """
    sorted_commands = commands.keys()
    sorted_commands.sort()
    return sorted_commands

def show_commands(commands, group=None):
    """
        Show commands within a group (or at the top level).
    """
    sorted_commands = get_sorted_commands(commands)

    for command in sorted_commands:
        if not commands[command].has_key('function'):
            continue

        option = command.replace('_','-')
        if group:
            prefix = "%-4s%-21s" % ('', option)
        else:
            prefix = "%-25s" % option

        if commands[command]['description'] is not None:
            print "%s - %s" % (prefix, commands[command]['description'])
        else:
            print prefix

    for command in sorted_commands:
        if not commands[command].has_key('function'):
            # This is a nested command
            print "\n" + _("Command Group: %s") % (command) + "\n"
            show_commands(commands[command], command)

def import_command_options(name, command_key, package, prefix, commands):
    """
        Obtain the options for the given command.
    """
    try:
        command_name = commands[name][command_key]
        if commands[name].has_key('group'):
            group = commands[name]['group']
            cmd = import_object("%s.%s.%s%s" % (package, group, prefix, command_name), "cli_options")
        else:
            cmd = import_object("%s.%s%s" % (package, prefix, command_name), "cli_options")
        cmd()
    except ImportError, errmsg:
        pass

def import_commands(dirname, package, fileprefix, commands):
    """
        Import commands from the given directory root for the given package and
        involving modules with the given prefix, initialising the commands
        structure.
    """
    for filename in os.listdir(dirname):
        pathname = join(dirname, filename)
        if isfile(pathname) and filename.startswith(fileprefix) and filename.endswith('.py'):
            module_name = filename.replace('.py','')
            cmd = import_object("%s.%s" % (package, module_name), "__init__")
            cmd()
        elif isdir(pathname):
            register_group(dirname, package, filename, fileprefix, commands)

def register_command(commands, command_key, name, func, group=None, description=None, aliases=[], after=[], before=[]):
    if group is not None:
        command = "%s_%s" % (group,name)
    else:
        command = name

    if isinstance(aliases, basestring):
        aliases = [aliases]

    if commands.has_key(command):
        log.fatal(_("Command '%s' already registered") % (command))
        sys.exit(1)

    if callable(func):
        if group is None:
            commands[name] = {
                    command_key: name,
                    'function': func,
                    'description': description,
                    'after': after,
                    'before': before,
                }
        else:
            commands[group][name] = {
                    'function': func,
                    'description': description,
                    'after': after,
                    'before': before,
                }

            commands[command] = commands[group][name]
            commands[command]['group'] = group
            commands[command][command_key] = name

        for alias in aliases:
            commands[alias] = {
                    command_key: name,
                    'function': func,
                    'description': _("Alias for %s") % name.replace('_', '-')
                }

def register_group(dirname, package, filename, fileprefix, commands):
    commands[filename] = {}
    import_commands(join(dirname, filename), "%s.%s" % (package, filename), fileprefix, commands)

def not_yet_implemented(*args, **kw):
    print _("Not yet implemented")
    sys.exit(1)
