# -*- coding: utf-8 -*-
# Copyright 2010-2013 Kolab Systems AG (http://www.kolabsys.com)
# Copyright 2013, 2014 Paul Boddie <paul@boddie.org.uk>
#
# Jeroen van Meeuwen (Kolab Systems) <vanmeeuwen a kolabsys.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

from glob import glob
from os.path import join
from subprocess import call, Popen, PIPE
import hashlib
import os
import random
import re
import sys
import time

import components

import pykolab

from pykolab import utils
from pykolab.constants import *
from pykolab.setup.services import *
from pykolab.translate import _

log = pykolab.getLogger('pykolab.setup')
conf = pykolab.getConf()

def __init__():
    components.register('roundcube', execute, description=description(), after=['mysql','ldap'])

def cli_options():
    roundcube_group = conf.add_cli_parser_option_group(_("Roundcube Options"))

    roundcube_group.add_option(
            "--reset-roundcube-config",
            dest    = "reset_roundcube_config",
            action  = "store_true",
            default = False,
            help    = _("Reset the Roundcube configuration.")
        )

def description():
    return _("Setup Roundcube.")

def random_digest():
    return hashlib.md5("%s" % random.random()).digest().encode("base64")

def execute(*args, **kw):

    # Signal that interaction may occur. This will involve debconf and similar
    # system-specific mechanisms if available.

    start_interaction("kolab-conf/title-roundcube") 
    try:
        _execute(*args, **kw)
    finally:
        stop_interaction()

def _execute(*args, **kw):

    # Stop if MySQL is not actually installed.

    if not have_mysql():
        if conf.check_only:
            utils.setup_status("roundcube", _("needs setup"))
            return
        else:
            print >> sys.stderr, _("MySQL not installed: cannot initialise Roundcube.")
            sys.exit(1)

    # Obtain an SMTP server.

    smtp_host = conf.get('smtp', 'host')

    if not smtp_host:
        if conf.check_only:
            utils.setup_status("roundcube", _("needs setup"))
            return

        smtp_host = ask_question("kolab-conf/smtp-server-selection",
                _("""
                    Please indicate the SMTP server to be used by the Kolab
                    components. (If this is a remote server, no attempt will
                    be made to set up server instances or to configure
                    services.)
                    """),
                _("SMTP server host"),
                default="localhost"
            )

    # Use the existing root credentials.

    defaults_file = get_mysql_defaults()

    if not isfile(defaults_file):
        if conf.check_only:
            utils.setup_status("roundcube", _("needs setup"))
            return
        make_mysql_defaults_file(defaults_file)

    # Test for a MySQL database.

    if not have_mysql_database(defaults_file, 'roundcube'):
        if conf.check_only:
            utils.setup_status("roundcube", _("needs setup"))
            return

        # Gather schema files for database initialisation.

        rcpath = get_roundcube_resources()

        if not rcpath:
            log.error(_("Roundcube installation path not found."))
            return

        # Retain the ordering.

        schema_files = []

        for roundcube_dir in glob('/usr/share/doc/roundcubemail*'):
            for root, directories, filenames in os.walk(roundcube_dir):
                for filename in filenames:
                    if filename.startswith('mysql.initial') and filename.endswith('.sql'):
                        schema_files.append(join(root,filename))

        for search_root in [
            join(rcpath, 'plugins/calendar/drivers/kolab/'),
            join(rcpath, 'plugins/libkolab/')
            ]:

            for root, directories, filenames in os.walk(search_root):
                for filename in filenames:
                    if filename.startswith('mysql') and filename.endswith('.sql'):
                        schema_files.append(join(root,filename))

        # Create the database.

        call(['mysql', '--defaults-file=%s' % defaults_file, '-e', 'create database roundcube'])

        # Apply each schema file.
        # MySQL does not have a read-from-file option as such.

        for schema_file in schema_files:
            p1 = Popen(['cat', schema_file], stdout=PIPE)
            p2 = Popen(['mysql', '--defaults-file=%s' % defaults_file, 'roundcube'], stdin=p1.stdout)
            p2.communicate()

    # Test for a MySQL user. Reset the user and configuration if explicitly
    # requested.

    if have_mysql_user(defaults_file, 'roundcube') and not conf.reset_roundcube_config:
        if not conf.check_only:
            print >> sys.stderr, _("Roundcube database account already exists and will not be configured.")
            log.info(_("A user called %s already exists. Not creating another one.") % 'roundcube')
    else:
        if conf.check_only:
            utils.setup_status("roundcube", _("needs setup"))
            return

        mysql_roundcube_password = ask_question("kolab-conf/roundcube-password",
                _("""
                    Please supply a password for the MySQL user 'roundcube'.
                    This password will be used by the Roundcube webmail
                    interface.
                    """),
                _("MySQL roundcube password"),
                default=utils.generate_password(),
                password=True,
                confirm=True
            )

        conf.mysql_roundcube_password = mysql_roundcube_password

        # Send the password details via a pipe to avoid leakage via the process list.

        p1 = Popen(['echo', "grant all privileges on roundcube.* to 'roundcube'@'localhost' identified by '%s'" %
            mysql_roundcube_password.replace("'", "''")], stdout=PIPE)
        p2 = Popen(['mysql', '--defaults-file=%s' % defaults_file], stdin=p1.stdout)
        p2.communicate()

        call(['mysql', '--defaults-file=%s' % defaults_file, '-e', 'flush privileges'])

        # Write out the settings including the password details.

        rc_settings = {
                'des_key': re.sub(
                        r'[^a-zA-Z0-9]',
                        "",
                        "%s%s" % (random_digest(), random_digest())
                    )[:24],

                'imap_admin_login': conf.get('cyrus-imap', 'admin_login'),
                'imap_admin_password': conf.get('cyrus-imap', 'admin_password'),
                'ldap_base_dn': conf.get('ldap', 'base_dn'),
                'ldap_group_base_dn': conf.get('ldap', 'group_base_dn'),
                'ldap_group_filter': conf.get('ldap', 'group_filter'),
                'ldap_ldap_uri': conf.get('ldap', 'ldap_uri'),
                'ldap_service_bind_dn': conf.get('ldap', 'service_bind_dn'),
                'ldap_service_bind_pw': conf.get('ldap', 'service_bind_pw'),
                'ldap_user_base_dn': conf.get('ldap', 'user_base_dn'),
                'ldap_user_filter': conf.get('ldap', 'user_filter'),
                'primary_domain': conf.get('kolab','primary_domain'),
                'mysql_uri': 'mysqli://roundcube:%s@localhost/roundcube' % (mysql_roundcube_password),
                'ldap_host': get_host_from_url(conf.get('ldap', 'ldap_uri')),
                'imap_host': get_host_from_url(conf.get('cyrus-imap', 'uri')),
                'smtp_host': smtp_host,
                'conf': conf,
            }

        want_files = [
                'acl.inc.php',
                'calendar.inc.php',
                'config.inc.php',
                'kolab_auth.inc.php',
                'kolab_delegation.inc.php',
                'kolab_files.inc.php',
                'kolab_folders.inc.php',
                'libkolab.inc.php',
                'managesieve.inc.php',
                'owncloud.inc.php',
                'password.inc.php',
                'recipient_to_contact.inc.php',
                'terms.html',
                'terms.inc.php'
            ]

        for want_file in want_files:
            template_file = get_template_path('roundcubemail/%s.tpl' % want_file)
            roundcube_dir = get_roundcube_config()

            if roundcube_dir:
                output_file = join(roundcube_dir, want_file)
            else:
                output_file = None

            if template_file is not None and output_file is not None:
                log.debug(_("Using template file %r") % (template_file), level=8)
                instantiate_template(template_file, output_file, [rc_settings])

    # If nothing needed updating, assume that the setup was done.

    if conf.check_only:
        utils.setup_status("roundcube", _("setup done"))

    else:
        # Update the configuration file.

        if not conf.has_section('smtp'):
            conf.add_section('smtp')
        conf.command_set('smtp', 'host', smtp_host)

        fp = open(conf.defaults.config_file, "w+")
        conf.cfg_parser.write(fp)
        fp.close()

        # Restart the Web server.

        time.sleep(2)

        apache = is_debian() and "apache2" or "httpd"
        if not control_service(apache, 'restart'):
            log.error(_("Could not start the webserver service."))

        if not configure_service(apache, True):
            log.error(_("Could not configure to start on boot, the " + \
                    "webserver service."))
